

/* Note from Bob Morein: This is a "pre-release" version of Simon's */
/* software. Simon is using VML PROLOG to develop this system. It   */
/* hasn't been checked out yet on PD and ED PROLOG.                 */



/* ATTENTION:                                                      */
/* This program is placed in the public domain on the condition    */
/* that the user in no way uses all or part of it in a profit      */
/* making resale capacity.
/*                                                                 */
/* Copyright: Simon Y. Blackwell 1985                              */
/*                                                                 */
/* Address comments and suggestions to:                            */
/*                                                                 */
/*                     Simon Y. Blackwell                          */
/*                     Dept. of Philosophy                         */
/*                     Bowling Green State Univ.                   */
/*                     Bowling Green, OH  43403                    */
/*                                                                 */
/*                                                                 */
/*   PIE.TM : A PROLOG INFERENCE ENGINE AND TRUTH MAINTENANCE      */
/*            SYSTEM                                               */



/* This file contains most of the fundamental predicates necessary */
/* for doing truth maintenance. PIE uses the prolog interpreter as */
/* an input parser by declaring most of the PIE syntax as goals.   */
/* Prior to execution the operators must be declared and the file  */
/* 'gensym' must be consulted. This is simplified by using the     */
/* redirect feature of ADA Prolog with the command line:           */
/*                     'prolog kops'                               */
/* The system is not yet complete and several extentions are       */
/* planned, many of which have already been implemented but        */
/* remain to be integrated with this particular piece of code.     */
/* Examples of planned extentions follow: one-directional rules,   */
/* a non-rule based inference based on mathematical set covering,  */
/* confidence factors, and more refined techniques for displaying  */
/* and editing a knowledge base. At the moment it is useful to know*/
/* or have a copy of the underlying representation. There is not   */
/* a lot of code here and it has not been thoroughly tested, but it*/
/* is quite powerful and flexible.                                 */


/* Sets 'X implies Y' up as a goal. NOTE: In order for the input   */
/* be parsed properly antecedents and consequents must be given as */
/* lists, e.g. '[X is a male,X is a human] implies [X is a man]'.  */
/* Consequents may themselves be rule declarations. The rules    */
/* are bi-directional and may contain Prolog goals as elements  */
/* of the antecedent or consequent lists. To force forward      */
/* chaining 'fc' may be made a member of the antecednt or       */
/* consequent lists.                                            */

X implies Y :-
	assert_r(X implies Y).



/* This is a special instance of the PIE assert. It allows new  */
/* relations to be declared in the form of operators. Asserting */
/* 'loves is a relation' will allow subsequent use of 'loves' as*/
/* an infix operator in antecedents or consequents of rules,    */
/* e.g. [X loves Y] implies [Y loves X].                        */

assert(X is a relation) :-
	gensym(rel,N),
	asserta(relation(N,X)),
	op(10,xfx,X).



/*The PIE assert adds facts to the knowledge base. While doing  */
/*so it checks to make sure that no conflicting facts exist. If */
/*conflicting facts do exist their identity is displayed.       */
/*Planned extentions include backward truth maintenance, wherein */
/*the inferences that led to both of the conflicting facts will */
/*be evaluated for confidence and 'distance' from input.        */
/* A typical assertion made by the user might look like:       */
/*                assert(bill is a man).                       */
/* If the assert(X) is followed by an 'fc', forward chaining  */
/* will occur for the entire system.                          */

assert(X):-
	fact(Number,Dependence,X,Confidence).
assert(X):-
	fact(Number,Dependence,not(X),Confidence),
	print('Sorry, in conflict with existing information.'),nl,
	prt_dependency(Number).
assert(not(X)):-
	fact(Number,Dependence,X,Confidence),
	print('Sorry, in conflict with existing information.'),nl,
	prt_dependency(Number).
assert(X):-
	functor(X,F,N),
	(atom(X);N>0),
	gensym(f,Number),
	assertz(fact(Number,input,X,Conf)),!,
	print('Inserted:'),write(Number),tab(1),write(X),nl.

/* Specifically designed for adding rules to the knowledge base */

assert_r(X):-
	functor(X,F,N),
	F=implies,
	gensym(r,Number),
	assertz(rule(Number,input,X,Conf)),!,
	print('Inserted:'),write(Number),tab(1),write(X),nl.


/* Cycles through all the forward chaining rules to find out if  */
/* the most recent assertion will cause any to fire. The         */
/* efficiency of this function can be increased dramatically by  */
/* copying the original rule to a 'non-conflict' stack and       */
/* effacing those conditions that have already been met. This    */
/* will result in ever shorter antecednt lists for the rules.    */


fc:-
	clause(fact(_,_,X,_),true),
	clause(rule(N,D,Y implies Z,C),true),
	given_member(X,Y),
	check_mult_con(N,Z),
	fail.
fc.

/* Checks to see if an antecedent that is part of a list exists  */
/* as a given in the kb.                                         */

given_member(X,[X|Y]):- Y=[].
given_member(X,[Y|Z]):-
	(Y;fact(N,D,Y,C)),
	given_member(X,Z).

/* Reads through a list of consequents and passes them on to     */
/* the infer function only if they do not already exist in the   */
/* kb. This should be enhanced so that confidence factors can    */
/* be incremented.                                               */

check_mult_con(N,[]).
check_mult_con(N,[X|Y]):-
	infer(N,X),
	check_mult_con(Y).

/* The 'infer' clause allows assertions to be made as a result of */
/* inference. It is similar to 'assert', but allows the passing   */
/* of a dependency bound to 'N'.                                  */

infer(N,X):-
	fact(Number,Dependence,not(X),Confidence),
	print('Sorry, in conflict with existing information.'),nl,
	prt_dependency(Number).
infer(N,X):-
	(X;fact(Num,Dep,X,Conf);rule(Num,Dep,X,Conf)).
infer(N,X):-
	functor(X,F,N1),
	F='implies',
	gensym(r,Number),
	assertz(rule(Number,N,X,Conf)),!,
	print('Inserted:'),write(Number),tab(1),write(X).
infer(N,X):-
	(atom(X);true),
	gensym(f,Number),
	assertz(fact(Number,N,X,Conf)),!,
	print('Inserted:'),write(Number),tab(1),write(X),nl.

/* A simple recursive function that will print out the rule    */
/* numbers on which a fact or rule depends. Extensions to this */
/* will allow for viewing in various modes and editing.        */

prt_dependency(N):-
	(fact(N,D,X,Conf);rule(N,D,X,Conf)),
	write(D),put(44),
	prt_dependency(D).
prt_dependency(N).

/* Allows removal of rules or facts by reference to their gensym  */
/* index. This could easily be enhanced by allowing instantiation */
/* through explicitly typing out the item to be removed.          */
/* It will automatically ensure that any inferences made on the   */
/* base of the retracted  item are also retracted!                */

remove(N):-
     display( 'remove: ', rule( N,D,X implies Y,C)),
	retract(rule(N,D,X implies Y,C)),
	print('Removed:'),write(N),tab(1),write(X),print('implies'),write(Y),nl,
	remove_con(N,Y).
remove(N):-
	retract(fact(N,D,X,C)),
	clause(rule(N1,_,Y implies Z,_),true),
	given_member(X,Y),
	print('Removed:'),write(N),tab(1),write(X),nl,
	remove_con(N,Z),
	fail.

/* 'Remove' will automatically forward chain in order re-infer  */
/* things that may be obtained through a different route than   */
/* that affected by the retraction process. This is necessary   */
/* because not all facts are taken advantage of in inferencing. */
/* That is to say, if a fact already exists 'infer' and 'assert'*/
/* will not add them redundantly to the kb. This will change    */
/* with the addition of confidence factors.                     */

remove(N):-
	fc.


/* Exhaustively checks facts in the kb and removes them if they */
/* depend on another item removed. NOTE: 'N=D' is part of a     */
/* disjunction, if it fails the fact will be reinserted in the  */
/* kb. At the moment this does not take advantage of the ADA    */
/* Prolog indexing capability, but it should in a dedicated     */
/* ADA application.                                             */

remove_con(N,[]).
remove_con(N,[X|Y]):-
	retract(fact(N1,D,X,C)),
	(N=D,print('Removed:'),write(N1),tab(1),write(X),nl);
	assertz(fact(N1,D,X,C)),
	remove_con(Y).
remove_con([X|Y]):-
	remove_con(Y).

/* Standard Prolog append.                                      */

append([],X,X).
append([A|B],C,[A|D]):-
	append(B,C,D).


/* Standard Prolog member.                                      */

mem(X,[X|_]).
mem(X,[Y|Z]):-
	mem(X,Z).


/* Activates backward chaining. A complex function, the first */
/* two clauses REQUIRE a list to function properly, but valid- */
/* ation is not done. This is required by the inference       */
/* mechanism. Its effect is to ensure that inheritance is not */
/* carried over to uninstantiated objects.                    */

obtain([]).
obtain(X):-
	X =[Y|Z],
	obtain_1(Y),
	obtain(Z).
obtain_1(X):-
	X.
obtain_1(X):-
	fact(N,D,X,C).
obtain_1(X):-
	clause(rule(N,D,Y implies Z,C),true),
	nl,
	mem(X,Z),
	obtain(Y),
	assert(X).
obtain_1(X):-!,
	ask(X).

/* At the moment user interaction with the system is limited*/
/* However, this author has several i/o utilities for menu    */
/* generation and input validation that have been tetsed in   */
/* a separate program.                                        */

ask(X):-
	print('Is it true that: '),
	write(X),nl,
	print('y/n/u/h:'),
	ratom(Y),
	check_response(Y,X).
check_response(y,X):-!,
	assert(X).
check_response(n,X):-!,
	print('No further inference possible.'),fail.
check_response(u,X):-!,
	print('Forward chaining to find other possible paths.'),nl,
	fc,obtain(X).
check_response(h,X):-!,
	nl,
	print('Upon full implementation problem specific help. For the time being make sure the argument you provided was in list form. y=yes, n=no, u=unknown, h=help.'),
	check_response_1(X).
check_response_1(X):-
	print('y/n/u:'),
	ratom(Y),
	check_response(Y,X).

